module Linear where

import Data.List
import Term
import Rule
import TRS
import SMT
import Proof


show_monomial :: (Int, Int) -> String
show_monomial (0, n) = show n
show_monomial (i, 1) = "x" ++ show i
show_monomial (i, n) = show n ++ " x" ++ show i

show_interpretation :: Interpretation -> String
show_interpretation (f, cs) =
  f ++ "_A(" ++ 
  intercalate "," [ show_monomial (i, 1) | i <- [1..length cs] ] ++ 
  ") = " ++
  intercalate " + " [ show_monomial (i, c) | (i, c) <- zip [0..] cs ]   

show_algebra :: Algebra -> String
show_algebra a =
  unlines [ show_interpretation interpretation | interpretation <- a ]

var_fi :: String -> Int -> Exp
var_fi f i = Var ("fi_" ++ f ++ "_" ++ show i) 

constant' :: Term -> [Exp]
constant' (V _) = []
constant' (F f ts) = 
  (var_fi f 0 :
   [ times [var_fi f i, ei]
   | (i, ti) <- zip [1..] ts,
     ei <- constant' ti ])

constant :: Term -> Exp
constant t = plus (constant' t)

coefficient' :: String -> Term -> [Exp]
coefficient' x (V y)
  | x == y    = [Val 1]
  | otherwise = []
coefficient' x (F f ts) = 
  [ times [var_fi f i, ei]
  | (i, ti) <- zip [1..] ts,
    ei <- coefficient' x ti ]

coefficient :: String -> Term -> Exp
coefficient x t = plus (coefficient' x t)

geqA :: Term -> Term -> Formula
geqA s t =
  conj (Geq (constant s) (constant t) :
        [ Geq (coefficient x s) (coefficient x t)
        | x <- Rule.variables (s, t) ])

gtA :: Term -> Term -> Formula
gtA s t =
  conj (Gt (constant s) (constant t) :
        [ Geq (coefficient x s) (coefficient x t)
        | x <- Rule.variables (s, t) ])

encode :: Signature -> TRS -> TRS -> Formula
encode sig rs ss =
  conj
    ([ Geq (var_fi f 0) (Val 0) | (f, _) <- sig ] ++
     [ Gt (var_fi f i) (Val 0) | (f, n) <- sig, i <- [1..n] ] ++  
     [ gtA  l r | (l, r) <- rs ] ++
     [ geqA l r | (l, r) <- ss ])

decode_interpretation :: Model -> String -> Int -> Interpretation
decode_interpretation model f n =
  (f, [ eval_exp model (var_fi f i) | i <- [0..n] ])

decode_algebra :: Model -> Signature -> Algebra
decode_algebra model sig =
  [ decode_interpretation model f n | (f, n) <- sig ] 

terminating :: String -> TRS -> TRS -> IO (Maybe SNProof) 
terminating _   [] _  = return (Just RIsEmpty)
terminating smt rs ss = do
  m <- sat smt (encode sig rs ss)
  case m of
    Nothing    -> return Nothing
    Just model -> return (Just (Linear (decode_algebra model sig)))
  where
    sig = TRS.signature_of (rs ++ ss)