module NonConfluenceCert where

import Term
import Rule
import TRS
import Proof
import CeTA


criticalPeaks :: TRS -> [Fork]
criticalPeaks trs =
  [ (t, rule1, p, s, rule2, u)
  | rule2 <- trs,
    let (l2, r2) = Rule.rename "y" 1 rule2,
    p <- positions l2,
    let l2p = subterm_at l2 p,
    rule1 <- trs,
    let (l1, r1) = Rule.rename "x" 1 rule1,
    Just sigma <- [mgu l1 l2p],
    let s = Term.substitute l2 sigma,
    let t = Term.substitute (replace l2 r1 p) sigma,
    let u = Term.substitute r2 sigma ]

tcap' :: String -> Int -> TRS -> Term -> (Int, Term)
tcap' x k _   (V _) = (k + 1, V (x ++ show k))
tcap' x k trs (F f ts)
  | any (\(l, _) -> unifiable u l) trs = (k + 1, V (x ++ show k))
  | otherwise = (m, u)
  where 
    (m, us) = tcapList x k trs ts
    u = F f us

tcapList :: String -> Int -> TRS -> [Term] -> (Int, [Term])
tcapList _ k _   []       = (k, [])
tcapList x k trs (t : ts) = (n, u : us)
    where 
      (m, u)  = tcap' x k trs t
      (n, us) = tcapList x m trs ts

tcap :: String -> TRS -> Term -> Term
tcap x trs t = u
  where (_, u) = tcap' x 0 trs t

groundSubstitution :: Term -> Subst
groundSubstitution t = [(v, F v []) | v <- Term.variables t]

notJoinable :: TRS -> Term -> Term -> Bool
notJoinable trs t u = not (unifiable (tcap "x" trs t) (tcap "y" trs u))

nonJoinableForks :: TRS -> [Fork]
nonJoinableForks trs =
  [ (t_sigma, rule1, p, s_sigma, rule2, u_sigma) 
  | (t, rule1, p, s, rule2, u) <- criticalPeaks trs, 
    let sigma = groundSubstitution s,
    let t_sigma = Term.substitute t sigma,
    let u_sigma = Term.substitute u sigma,
    notJoinable trs t_sigma u_sigma,
    let s_sigma = Term.substitute s sigma ]

disprove :: Criterion
disprove trsR =
  case nonJoinableForks trsR of
    []       -> return Nothing
    fork : _ -> return (Just (CRDisproof (CapNotUnif fork)))
